import optparse

from ._compat import Command, cmdoptions


class PipCommand(Command):
    name = "PipCommand"


def get_pip_command():
    # Use pip's parser for pip.conf management and defaults.
    # General options (find_links, index_url, extra_index_url, trusted_host,
    # and pre) are defered to pip.
    pip_command = PipCommand()
    pip_command.parser.add_option(cmdoptions.no_binary())
    pip_command.parser.add_option(cmdoptions.only_binary())
    index_opts = cmdoptions.make_option_group(
        cmdoptions.index_group, pip_command.parser
    )
    pip_command.parser.insert_option_group(0, index_opts)
    pip_command.parser.add_option(
        optparse.Option("--pre", action="store_true", default=False)
    )

    return pip_command


pip_command = get_pip_command()

# Get default values of the pip's options (including options from pip.conf).
pip_defaults = pip_command.parser.get_default_values()



import os
import sys
from itertools import chain

from .click import get_os_args, unstyle
from .logging import log
from .utils import UNSAFE_PACKAGES, comment, dedup, format_requirement, key_from_req


class OutputWriter(object):
    def __init__(
        self,
        src_files,
        dst_file,
        dry_run,
        emit_header,
        emit_index,
        emit_trusted_host,
        annotate,
        generate_hashes,
        default_index_url,
        index_urls,
        trusted_hosts,
        format_control,
        allow_unsafe,
    ):
        self.src_files = src_files
        self.dst_file = dst_file
        self.dry_run = dry_run
        self.emit_header = emit_header
        self.emit_index = emit_index
        self.emit_trusted_host = emit_trusted_host
        self.annotate = annotate
        self.generate_hashes = generate_hashes
        self.default_index_url = default_index_url
        self.index_urls = index_urls
        self.trusted_hosts = trusted_hosts
        self.format_control = format_control
        self.allow_unsafe = allow_unsafe

    def _sort_key(self, ireq):
        return (not ireq.editable, str(ireq.req).lower())

    def write_header(self):
        if self.emit_header:
            yield comment("#")
            yield comment("# This file is autogenerated by pip-compile")
            yield comment("# To update, run:")
            yield comment("#")
            custom_cmd = os.environ.get("CUSTOM_COMPILE_COMMAND")
            if custom_cmd:
                yield comment("#    {}".format(custom_cmd))
            else:
                prog = os.path.basename(sys.argv[0])
                args = " ".join(get_os_args())
                yield comment("#    {prog} {args}".format(prog=prog, args=args))
            yield comment("#")

    def write_index_options(self):
        if self.emit_index:
            for index, index_url in enumerate(dedup(self.index_urls)):
                if index_url.rstrip("/") == self.default_index_url:
                    continue
                flag = "--index-url" if index == 0 else "--extra-index-url"
                yield "{} {}".format(flag, index_url)

    def write_trusted_hosts(self):
        if self.emit_trusted_host:
            for trusted_host in dedup(self.trusted_hosts):
                yield "--trusted-host {}".format(trusted_host)

    def write_format_controls(self):
        for nb in dedup(self.format_control.no_binary):
            yield "--no-binary {}".format(nb)
        for ob in dedup(self.format_control.only_binary):
            yield "--only-binary {}".format(ob)

    def write_flags(self):
        emitted = False
        for line in chain(
            self.write_index_options(),
            self.write_trusted_hosts(),
            self.write_format_controls(),
        ):
            emitted = True
            yield line
        if emitted:
            yield ""

    def _iter_lines(
        self,
        results,
        unsafe_requirements,
        reverse_dependencies,
        primary_packages,
        markers,
        hashes,
    ):
        for line in self.write_header():
            yield line
        for line in self.write_flags():
            yield line

        unsafe_requirements = (
            {r for r in results if r.name in UNSAFE_PACKAGES}
            if not unsafe_requirements
            else unsafe_requirements
        )  # noqa
        packages = {r for r in results if r.name not in UNSAFE_PACKAGES}

        packages = sorted(packages, key=self._sort_key)

        for ireq in packages:
            line = self._format_requirement(
                ireq,
                reverse_dependencies,
                primary_packages,
                markers.get(key_from_req(ireq.req)),
                hashes=hashes,
            )
            yield line

        if unsafe_requirements:
            unsafe_requirements = sorted(unsafe_requirements, key=self._sort_key)
            yield ""
            yield comment(
                "# The following packages are considered "
                "to be unsafe in a requirements file:"
            )

            for ireq in unsafe_requirements:
                req = self._format_requirement(
                    ireq,
                    reverse_dependencies,
                    primary_packages,
                    marker=markers.get(key_from_req(ireq.req)),
                    hashes=hashes,
                )
                if not self.allow_unsafe:
                    yield comment("# {}".format(req))
                else:
                    yield req

    def write(
        self,
        results,
        unsafe_requirements,
        reverse_dependencies,
        primary_packages,
        markers,
        hashes,
    ):

        for line in self._iter_lines(
            results,
            unsafe_requirements,
            reverse_dependencies,
            primary_packages,
            markers,
            hashes,
        ):
            log.info(line)
            if not self.dry_run:
                self.dst_file.write(unstyle(line).encode("utf-8"))
                self.dst_file.write(os.linesep.encode("utf-8"))

    def _format_requirement(
        self, ireq, reverse_dependencies, primary_packages, marker=None, hashes=None
    ):
        ireq_hashes = (hashes if hashes is not None else {}).get(ireq)

        line = format_requirement(ireq, marker=marker, hashes=ireq_hashes)

        if not self.annotate or key_from_req(ireq.req) in primary_packages:
            return line

        # Annotate what packages this package is required by
        required_by = reverse_dependencies.get(ireq.name.lower(), [])
        if required_by:
            annotation = ", ".join(sorted(required_by))
            line = "{:24}{}{}".format(
                line,
                " \\\n    " if ireq_hashes else "  ",
                comment("# via " + annotation),
            )
        return line



from pytest import mark, raises

from piptools.utils import (
    as_tuple,
    dedup,
    flat_map,
    format_requirement,
    format_specifier,
    get_hashes_from_ireq,
    is_pinned_requirement,
    name_from_req,
)


def test_format_requirement(from_line):
    ireq = from_line("test==1.2")
    assert format_requirement(ireq) == "test==1.2"


def test_format_requirement_editable(from_editable):
    ireq = from_editable("git+git://fake.org/x/y.git#egg=y")
    assert format_requirement(ireq) == "-e git+git://fake.org/x/y.git#egg=y"


def test_format_requirement_ireq_with_hashes(from_line):
    ireq = from_line("pytz==2017.2")
    ireq_hashes = [
        "sha256:d1d6729c85acea5423671382868627129432fba9a89ecbb248d8d1c7a9f01c67",
        "sha256:f5c056e8f62d45ba8215e5cb8f50dfccb198b4b9fbea8500674f3443e4689589",
    ]

    expected = (
        "pytz==2017.2 \\\n"
        "    --hash=sha256:d1d6729c85acea542367138286"
        "8627129432fba9a89ecbb248d8d1c7a9f01c67 \\\n"
        "    --hash=sha256:f5c056e8f62d45ba8215e5cb8f5"
        "0dfccb198b4b9fbea8500674f3443e4689589"
    )
    assert format_requirement(ireq, hashes=ireq_hashes) == expected


def test_format_requirement_ireq_with_hashes_and_markers(from_line):
    ireq = from_line("pytz==2017.2")
    marker = 'python_version<"3.0"'
    ireq_hashes = [
        "sha256:d1d6729c85acea5423671382868627129432fba9a89ecbb248d8d1c7a9f01c67",
        "sha256:f5c056e8f62d45ba8215e5cb8f50dfccb198b4b9fbea8500674f3443e4689589",
    ]

    expected = (
        'pytz==2017.2 ; python_version<"3.0" \\\n'
        "    --hash=sha256:d1d6729c85acea542367138286"
        "8627129432fba9a89ecbb248d8d1c7a9f01c67 \\\n"
        "    --hash=sha256:f5c056e8f62d45ba8215e5cb8f5"
        "0dfccb198b4b9fbea8500674f3443e4689589"
    )
    assert format_requirement(ireq, marker, hashes=ireq_hashes) == expected


def test_format_specifier(from_line):
    ireq = from_line("foo")
    assert format_specifier(ireq) == "<any>"

    ireq = from_line("foo==1.2")
    assert format_specifier(ireq) == "==1.2"

    ireq = from_line("foo>1.2,~=1.1,<1.5")
    assert format_specifier(ireq) == "~=1.1,>1.2,<1.5"
    ireq = from_line("foo~=1.1,<1.5,>1.2")
    assert format_specifier(ireq) == "~=1.1,>1.2,<1.5"


def test_as_tuple(from_line):
    ireq = from_line("foo==1.1")
    name, version, extras = as_tuple(ireq)
    assert name == "foo"
    assert version == "1.1"
    assert extras == ()

    ireq = from_line("foo[extra1,extra2]==1.1")
    name, version, extras = as_tuple(ireq)
    assert name == "foo"
    assert version == "1.1"
    assert extras == ("extra1", "extra2")

    # Non-pinned versions aren't accepted
    should_be_rejected = ["foo==1.*", "foo~=1.1,<1.5,>1.2", "foo"]
    for spec in should_be_rejected:
        ireq = from_line(spec)
        with raises(TypeError):
            as_tuple(ireq)


def test_flat_map():
    assert [1, 2, 4, 1, 3, 9] == list(flat_map(lambda x: [1, x, x * x], [2, 3]))


def test_dedup():
    assert list(dedup([3, 1, 2, 4, 3, 5])) == [3, 1, 2, 4, 5]


def test_get_hashes_from_ireq(from_line):
    ireq = from_line(
        "pytz==2017.2",
        options={
            "hashes": {
                "sha256": [
                    "d1d6729c85acea5423671382868627129432fba9a89ecbb248d8d1c7a9f01c67",
                    "f5c056e8f62d45ba8215e5cb8f50dfccb198b4b9fbea8500674f3443e4689589",
                ]
            }
        },
    )
    expected = [
        "sha256:d1d6729c85acea5423671382868627129432fba9a89ecbb248d8d1c7a9f01c67",
        "sha256:f5c056e8f62d45ba8215e5cb8f50dfccb198b4b9fbea8500674f3443e4689589",
    ]
    assert get_hashes_from_ireq(ireq) == expected


@mark.parametrize(
    ("line", "expected"),
    [
        ("django==1.8", True),
        ("django===1.8", True),
        ("django>1.8", False),
        ("django~=1.8", False),
        ("django==1.*", False),
    ],
)
def test_is_pinned_requirement(from_line, line, expected):
    ireq = from_line(line)
    assert is_pinned_requirement(ireq) is expected


def test_is_pinned_requirement_editable(from_editable):
    ireq = from_editable("git+git://fake.org/x/y.git#egg=y")
    assert not is_pinned_requirement(ireq)


def test_name_from_req(from_line):
    ireq = from_line("django==1.8")
    assert name_from_req(ireq.req) == "django"


def test_name_from_req_with_project_name(from_line):
    ireq = from_line("foo==1.8")
    ireq.req.project_name = "bar"
    assert name_from_req(ireq.req) == "bar"



import pytest

from piptools.exceptions import UnsupportedConstraint


@pytest.mark.parametrize(
    ("input", "expected", "prereleases"),
    (
        (tup + (False,))[:3]
        for tup in [
            (["Django"], ["django==1.8"]),
            (
                ["Flask"],
                [
                    "flask==0.10.1",
                    "itsdangerous==0.24",
                    "markupsafe==0.23",
                    "jinja2==2.7.3",
                    "werkzeug==0.10.4",
                ],
            ),
            (["Jinja2", "markupsafe"], ["jinja2==2.7.3", "markupsafe==0.23"]),
            # We should return a normal release version if prereleases is False
            (["SQLAlchemy"], ["sqlalchemy==0.9.9"]),
            # We should return the prerelease version if prereleases is True
            (["SQLAlchemy"], ["sqlalchemy==1.0.0b5"], True),
            # Ipython has extras available, but we don't require them in this test
            (["ipython"], ["ipython==2.1.0", "gnureadline==6.3.3"]),
            # We should get dependencies for extras
            (
                ["ipython[notebook]"],
                [
                    "ipython[notebook]==2.1.0",
                    "pyzmq==2.1.12",
                    "jinja2==2.7.3",
                    "tornado==3.2.2",
                    "markupsafe==0.23",
                    "gnureadline==6.3.3",
                ],
            ),
            # We should get dependencies for multiple extras
            (
                ["ipython[notebook,nbconvert]"],
                [
                    # Note that the extras should be sorted
                    "ipython[nbconvert,notebook]==2.1.0",
                    "pyzmq==2.1.12",
                    "jinja2==2.7.3",
                    "tornado==3.2.2",
                    "markupsafe==0.23",
                    "gnureadline==6.3.3",
                    "pygments==1.5",
                    "sphinx==0.3",
                ],
            ),
            # We must take the union of all extras
            (
                ["ipython[notebook]", "ipython[nbconvert]"],
                [
                    # Note that the extras should be sorted
                    "ipython[nbconvert,notebook]==2.1.0",
                    "pyzmq==2.1.12",
                    "jinja2==2.7.3",
                    "tornado==3.2.2",
                    "markupsafe==0.23",
                    "gnureadline==6.3.3",
                    "pygments==1.5",
                    "sphinx==0.3",
                ],
            ),
            # We must remove child dependencies from result if parent
            # is removed (e.g. vine from amqp>=2.0)
            # See: GH-370
            # because of upated dependencies in the test index, we need to pin celery
            # in order to reproduce vine removal (because it was readded in later
            # releases)
            (
                ["celery<=3.1.23", "librabbitmq"],
                [
                    "amqp==1.4.9",
                    "anyjson==0.3.3",
                    "billiard==3.5.0.2",
                    "celery==3.1.23",
                    "kombu==3.0.35",
                    "librabbitmq==1.6.1",
                    "pytz==2016.4",
                ],
            ),
            # Support specifying loose top-level requirements that could also appear as
            # pinned subdependencies.
            (
                ["billiard", "celery", "fake-piptools-test-with-pinned-deps"],
                [
                    "amqp==1.4.9",
                    "anyjson==0.3.3",
                    "billiard==3.3.0.23",
                    "celery==3.1.18",  # this is pinned from test subdependency
                    "fake-piptools-test-with-pinned-deps==0.1",
                    "kombu==3.0.35",
                    "pytz==2016.4",
                ],
            ),
            # Exclude package dependcy of setuptools as it is unsafe.
            (["html5lib"], ["html5lib==0.999999999"]),
            # We shouldn't include irrelevant pip constraints
            # See: GH-471
            (
                ["Flask", ("click", True), ("itsdangerous", True)],
                [
                    "flask==0.10.1",
                    "itsdangerous==0.24",
                    "markupsafe==0.23",
                    "jinja2==2.7.3",
                    "werkzeug==0.10.4",
                ],
            ),
            # Unsafe dependencies should be filtered
            (["setuptools==35.0.0", "anyjson==0.3.3"], ["anyjson==0.3.3"]),
            (
                ["fake-piptools-test-with-unsafe-deps==0.1"],
                ["fake-piptools-test-with-unsafe-deps==0.1"],
            ),
        ]
    ),
)
def test_resolver(resolver, from_line, input, expected, prereleases):
    input = [line if isinstance(line, tuple) else (line, False) for line in input]
    input = [from_line(req[0], constraint=req[1]) for req in input]
    output = resolver(input, prereleases=prereleases).resolve()
    output = {str(line) for line in output}
    assert output == {str(line) for line in expected}


@pytest.mark.parametrize(
    ("input", "expected", "prereleases"),
    (
        (tup + (False,))[:3]
        for tup in [
            (
                ["setuptools==34.0.0"],
                ["appdirs==1.4.9", "setuptools==34.0.0", "packaging==16.8"],
            ),
            (
                ["fake-piptools-test-with-unsafe-deps==0.1"],
                [
                    "appdirs==1.4.9",
                    "setuptools==34.0.0",
                    "packaging==16.8",
                    "fake-piptools-test-with-unsafe-deps==0.1",
                ],
            ),
        ]
    ),
)
def test_resolver__allows_unsafe_deps(
    resolver, from_line, input, expected, prereleases
):
    input = [line if isinstance(line, tuple) else (line, False) for line in input]
    input = [from_line(req[0], constraint=req[1]) for req in input]
    output = resolver(input, prereleases=prereleases, allow_unsafe=True).resolve()
    output = {str(line) for line in output}
    assert output == {str(line) for line in expected}


def test_resolver__max_number_rounds_reached(resolver, from_line):
    """
    Resolver should raise an exception if max round has been reached.
    """
    input = [from_line("django")]
    with pytest.raises(RuntimeError, match="after 0 rounds of resolving"):
        resolver(input).resolve(max_rounds=0)


def test_resolver__check_constraints(resolver, from_line):
    """
    Resolver should not support non-editable URLs as packages.
    """
    input = [from_line("django"), from_line("https://example.com/#egg=example")]
    with pytest.raises(
        UnsupportedConstraint, match="pip-compile does not support URLs as packages"
    ):
        resolver(input).resolve()


def test_iter_dependencies(resolver, from_line):
    """
    Dependencies should be pinned or editable.
    """
    ireq = from_line("django>=1.8")
    res = resolver([])

    with pytest.raises(
        TypeError, match="Expected pinned or editable requirement, got django>=1.8"
    ):
        next(res._iter_dependencies(ireq))



from pytest import fixture, mark, raises

from piptools._compat import FormatControl
from piptools.utils import comment
from piptools.writer import OutputWriter


@fixture
def writer(tmp_file):
    return OutputWriter(
        src_files=["src_file", "src_file2"],
        dst_file=tmp_file,
        dry_run=True,
        emit_header=True,
        emit_index=True,
        emit_trusted_host=True,
        annotate=True,
        generate_hashes=False,
        default_index_url=None,
        index_urls=[],
        trusted_hosts=[],
        format_control=FormatControl(set(), set()),
        allow_unsafe=False,
    )


def test_format_requirement_annotation_editable(from_editable, writer):
    # Annotations are printed as comments at a fixed column
    ireq = from_editable("git+git://fake.org/x/y.git#egg=y")
    reverse_dependencies = {"y": ["xyz"]}

    assert writer._format_requirement(
        ireq, reverse_dependencies, primary_packages=[]
    ) == "-e git+git://fake.org/x/y.git#egg=y  " + comment("# via xyz")


def test_format_requirement_annotation(from_line, writer):
    ireq = from_line("test==1.2")
    reverse_dependencies = {"test": ["xyz"]}

    assert writer._format_requirement(
        ireq, reverse_dependencies, primary_packages=[]
    ) == "test==1.2                 " + comment("# via xyz")


def test_format_requirement_annotation_lower_case(from_line, writer):
    ireq = from_line("Test==1.2")
    reverse_dependencies = {"test": ["xyz"]}

    assert writer._format_requirement(
        ireq, reverse_dependencies, primary_packages=[]
    ) == "test==1.2                 " + comment("# via xyz")


def test_format_requirement_not_for_primary(from_line, writer):
    "Primary packages should not get annotated."
    ireq = from_line("test==1.2")
    reverse_dependencies = {"test": ["xyz"]}

    assert (
        writer._format_requirement(
            ireq, reverse_dependencies, primary_packages=["test"]
        )
        == "test==1.2"
    )


def test_format_requirement_not_for_primary_lower_case(from_line, writer):
    "Primary packages should not get annotated."
    ireq = from_line("Test==1.2")
    reverse_dependencies = {"test": ["xyz"]}

    assert (
        writer._format_requirement(
            ireq, reverse_dependencies, primary_packages=["test"]
        )
        == "test==1.2"
    )


def test_format_requirement_environment_marker(from_line, writer):
    "Environment markers should get passed through to output."
    ireq = from_line(
        'test ; python_version == "2.7" and platform_python_implementation == "CPython"'
    )
    reverse_dependencies = set()

    result = writer._format_requirement(
        ireq, reverse_dependencies, primary_packages=["test"], marker=ireq.markers
    )
    assert (
        result == 'test ; python_version == "2.7" and '
        'platform_python_implementation == "CPython"'
    )


@mark.parametrize(("allow_unsafe",), [(True,), (False,)])
def test_iter_lines__unsafe_dependencies(writer, from_line, allow_unsafe):
    writer.allow_unsafe = allow_unsafe

    ireq = [from_line("test==1.2")]
    unsafe_req = [from_line("setuptools")]
    reverse_dependencies = {"test": ["xyz"]}

    str_lines = list(
        writer._iter_lines(ireq, unsafe_req, reverse_dependencies, ["test"], {}, None)
    )
    assert (
        comment(
            "# The following packages are considered "
            "to be unsafe in a requirements file:"
        )
        in str_lines
    )
    if allow_unsafe:
        assert "setuptools" in str_lines
    else:
        assert comment("# setuptools") in str_lines
    assert "test==1.2" in str_lines


def test_write_header(writer, monkeypatch):
    monkeypatch.setattr(
        "sys.argv",
        ["pip-compile", "--output-file", "dst_file", "src_file", "src_file2"],
    )
    expected = map(
        comment,
        [
            "#",
            "# This file is autogenerated by pip-compile",
            "# To update, run:",
            "#",
            "#    pip-compile --output-file dst_file src_file src_file2",
            "#",
        ],
    )
    assert list(writer.write_header()) == list(expected)


def test_write_header_custom_compile_command(writer, monkeypatch):
    monkeypatch.setenv("CUSTOM_COMPILE_COMMAND", "./pipcompilewrapper")
    expected = map(
        comment,
        [
            "#",
            "# This file is autogenerated by pip-compile",
            "# To update, run:",
            "#",
            "#    ./pipcompilewrapper",
            "#",
        ],
    )
    assert list(writer.write_header()) == list(expected)


def test_write_header_no_emit_header(writer):
    """
    There should not be headers if emit_header is False
    """
    writer.emit_header = False

    with raises(StopIteration):
        next(writer.write_header())


def test_write_format_controls(writer):
    """
    Tests --no-binary/--only-binary options.
    """

    writer.format_control = FormatControl(
        no_binary=["psycopg2", "click"], only_binary=["pytz", "django"]
    )
    lines = list(writer.write_format_controls())

    assert "--no-binary psycopg2" in lines
    assert "--no-binary click" in lines

    assert "--only-binary pytz" in lines
    assert "--only-binary django" in lines


@mark.parametrize(
    ("index_urls", "expected_lines"),
    (
        # No urls - no options
        [[], []],
        # Single URL should be index-url
        [["https://index-url.com"], ["--index-url https://index-url.com"]],
        # First URL should be index-url, the others should be extra-index-url
        [
            [
                "https://index-url1.com",
                "https://index-url2.com",
                "https://index-url3.com",
            ],
            [
                "--index-url https://index-url1.com",
                "--extra-index-url https://index-url2.com",
                "--extra-index-url https://index-url3.com",
            ],
        ],
        # If a first URL equals to the default URL, the the index url must not be set
        # and the others should be extra-index-url
        [
            [
                "https://default-index-url.com",
                "https://index-url1.com",
                "https://index-url2.com",
            ],
            [
                "--extra-index-url https://index-url1.com",
                "--extra-index-url https://index-url2.com",
            ],
        ],
        # Ignore URLs equal to the default index-url
        # (note: the previous case is exception)
        [
            [
                "https://index-url1.com",
                "https://default-index-url.com",
                "https://index-url2.com",
            ],
            [
                "--index-url https://index-url1.com",
                "--extra-index-url https://index-url2.com",
            ],
        ],
        # Ignore URLs equal to the default index-url
        [["https://default-index-url.com", "https://default-index-url.com"], []],
        # All URLs must be deduplicated
        [
            [
                "https://index-url1.com",
                "https://index-url1.com",
                "https://index-url2.com",
            ],
            [
                "--index-url https://index-url1.com",
                "--extra-index-url https://index-url2.com",
            ],
        ],
    ),
)
def test_write_index_options(writer, index_urls, expected_lines):
    """
    Test write_index_options method.
    """
    writer.index_urls = index_urls
    writer.default_index_url = "https://default-index-url.com"
    assert list(writer.write_index_options()) == expected_lines


def test_write_index_options_no_emit_index(writer):
    """
    There should not be --index-url/--extra-index-url options if emit_index is False.
    """
    writer.emit_index = False
    with raises(StopIteration):
        next(writer.write_index_options())
